package cat.hack3.codingtests.marsrover;

import cat.hack3.codingtests.marsrover.cartography.Coordinates;
import cat.hack3.codingtests.marsrover.cartography.Direction;
import cat.hack3.codingtests.marsrover.cartography.MarsMap;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.stream.IntStream;

import static cat.hack3.codingtests.marsrover.cartography.Direction.*;
import static org.testng.Assert.assertEquals;

public class MarsRoverTest {

    private MarsRover rover;
    private MoveForwardCommand moveForwardCommand;
    private MoveBackwardsCommand moveBackwardsCommand;
    private TurnLeftCommand turnLeftCommand;
    private TurnRightCommand turnRightCommand;

    @BeforeMethod
    public void setUp() {
        int mapWidth = 10;
        int mapHeight = 10;
        int latitudeStartingPoint = 2;
        int longitudeStartingPoint = 3;
        var startingCoordinates = Coordinates.of(latitudeStartingPoint, longitudeStartingPoint);
        var marsMap = new MarsMap(mapHeight, mapWidth, startingCoordinates);
        rover = new MarsRover(marsMap, SOUTH);

        moveForwardCommand = new MoveForwardCommand(rover);
        moveBackwardsCommand = new MoveBackwardsCommand(rover);
        turnLeftCommand = new TurnLeftCommand(rover);
        turnRightCommand = new TurnRightCommand(rover);
    }

    @Test
    public void stepForward() {
        Direction originalDirection = rover.getCurrentDirection();

        moveForwardCommand.execute();

        assertEquals(rover.getCurrentCoordinates(), Coordinates.of(3, 3));
        assertEquals(rover.getCurrentDirection(), originalDirection);
    }

    @Test
    public void stepBackwards() {
        Direction originalDirection = rover.getCurrentDirection();

        moveBackwardsCommand.execute();
        Coordinates currentPosition = rover.getCurrentCoordinates();

        assertEquals(currentPosition, Coordinates.of(1, 3));
        assertEquals(rover.getCurrentDirection(), originalDirection);
    }

    @Test
    public void turnLef() {
        turnLeftCommand.execute();
        assertEquals(rover.getCurrentDirection(), EAST);
        turnLeftCommand.execute();
        assertEquals(rover.getCurrentDirection(), NORTH);
        turnLeftCommand.execute();
        assertEquals(rover.getCurrentDirection(), WEST);
        turnLeftCommand.execute();
        assertEquals(rover.getCurrentDirection(), SOUTH);
        turnLeftCommand.execute();
        assertEquals(rover.getCurrentDirection(), EAST);
    }

    @Test
    public void turnRight() {
        turnRightCommand.execute();
        assertEquals(rover.getCurrentDirection(), WEST);
        turnRightCommand.execute();
        assertEquals(rover.getCurrentDirection(), NORTH);
        turnRightCommand.execute();
        assertEquals(rover.getCurrentDirection(), EAST);
        turnRightCommand.execute();
        assertEquals(rover.getCurrentDirection(), SOUTH);
        turnRightCommand.execute();
        assertEquals(rover.getCurrentDirection(), WEST);
    }

    @Test
    public void loopTheWorldInLatitudeMovingForward() {
        Coordinates originalCoordinates = rover.getCurrentCoordinates();
        Direction originalDirection = rover.getCurrentDirection();

        repeatAction(10, () -> moveForwardCommand.execute());

        assertEquals(rover.getCurrentCoordinates(), originalCoordinates);
        assertEquals(rover.getCurrentDirection(), originalDirection);
    }

    @Test
    public void loopTheWorldInLatitudeMovingBackwards() {
        Coordinates originalCoordinates = rover.getCurrentCoordinates();
        Direction originalDirection = rover.getCurrentDirection();

        repeatAction(10, () -> moveBackwardsCommand.execute());

        assertEquals(rover.getCurrentCoordinates(), originalCoordinates);
        assertEquals(rover.getCurrentDirection(), originalDirection);
    }

    @Test
    public void loopTheWorldInLongitudeMovingForward() {
        Coordinates originalCoordinates = rover.getCurrentCoordinates();

        turnLeftCommand.execute();
        repeatAction(10, () -> moveForwardCommand.execute());

        assertEquals(rover.getCurrentCoordinates(), originalCoordinates);
        assertEquals(rover.getCurrentDirection(), EAST);
    }

    @Test
    public void loopTheWorldInLongitudeMovingBackwards() {
        Coordinates originalCoordinates = rover.getCurrentCoordinates();

        turnLeftCommand.execute();
        repeatAction(10, () -> moveBackwardsCommand.execute());

        assertEquals(rover.getCurrentCoordinates(), originalCoordinates);
        assertEquals(rover.getCurrentDirection(), EAST);
    }

    @Test
    public void crazyTour() {
        moveForwardCommand.execute();
        assertEquals(rover.getCurrentCoordinates(), Coordinates.of(3, 3));
        assertEquals(rover.getCurrentDirection(), SOUTH);

        turnLeftCommand.execute();
        repeatAction(23, () -> moveForwardCommand.execute());
        assertEquals(rover.getCurrentCoordinates(), Coordinates.of(3, 6));
        assertEquals(rover.getCurrentDirection(), EAST);

        repeatAction(3, () -> turnRightCommand.execute());
        repeatAction(9, () -> moveForwardCommand.execute());
        assertEquals(rover.getCurrentCoordinates(), Coordinates.of(4, 6));
        assertEquals(rover.getCurrentDirection(), NORTH);

        moveBackwardsCommand.execute();
        turnRightCommand.execute();
        moveForwardCommand.execute();
        assertEquals(rover.getCurrentCoordinates(), Coordinates.of(5, 7));
        assertEquals(rover.getCurrentDirection(), EAST);
    }

    private void repeatAction(int times, Runnable actionToRepeat) {
        IntStream.rangeClosed(1, times)
                .forEach(i -> actionToRepeat.run());
    }
}
